import { createWriteStream, readFileSync, writeFileSync } from 'node:fs';
import { Readable } from 'node:stream';
import { fileFromSync } from 'fetch-blob/from.js';
import Blob from 'cross-blob';
import { Chunk, ChunkFile, ChunkType } from './gamefile/chunk_file.js';
import { Gamefile } from './gamefile/gamefile.js';
import { export_sector_map, import_sector_map } from './gamefile/sector_map.js';
import { ImageChunk } from './chunks/image.js';
import { GsColorParam, GsFilter, GsStorageFormat } from './ps2/gs_constants.js';
import { swizzle_32 } from './ps2/swizzle.js';
import { Material, MaterialPass, MaterialsChunk } from './chunks/materials.js';
import { export_gltf } from './gltf/gltf_export.js';
import { import_gltf } from './gltf/gltf_import.js';
import { PNG } from 'pngjs';
//let gamefile_blob = fileFromSync('D:\\dev\\ps2\\df\\GAMEFILE.DAT');
//let exe_blob = fileFromSync('D:\\dev\\ps2\\df\\SCUS_973.67');
let iso_blob = fileFromSync('D:\\dev\\ps2\\darkestfaerie.iso');
let gamefile_blob = iso_blob.slice(885373 * 2048, 885373 * 2048 + 1902385152);
let exe_blob = iso_blob.slice(390 * 2048, 390 * 2048 + 2291560);
(async () => {
    let sector_map = await import_sector_map(exe_blob);
    let gamefile = new Gamefile(gamefile_blob, sector_map);
    for (let [i, file] of [...gamefile].entries()) {
        if (i < 0x2a6)
            continue;
        try {
            console.log(i.toString(16) + ", " + i);
            let chunk_file = await ChunkFile.from_blob(file);
            let materials = await MaterialsChunk.from_blob(chunk_file.get_chunk_of_type(ChunkType.Materials).contents);
        }
        catch (e) {
            continue;
        }
    }
    return;
    /*let image_replacements = [
        [0x241, 'legal.data'],
        [0x258, 'neopets.data'],
        [0x249, 'title.data']
    ] as const;
    for(let [file, import_from] of image_replacements) {
        let chunk = await Chunk.from_blob(gamefile.get_file(file));
        let image = await ImageChunk.from_blob(chunk.contents);
        console.log(image.width + ", " + image.height);
        console.log(image.locations);

        image.import_data(image.locations[0], readFileSync(import_from));
        let clut_data = readFileSync(import_from + ".pal");
        let clut_full = new Uint8Array(1024);
        for(let i = 0; i < 256; i++) {
            clut_full[i*4] = clut_data[i*4+2];
            clut_full[i*4+1] = clut_data[i*4+1];
            clut_full[i*4+2] = clut_data[i*4];
            clut_full[i*4+3] = 0x80;
        }
        image.import_data(image.locations[1], clut_full);
        chunk.contents = image.to_blob();
        gamefile.replace_file(file, chunk.to_blob());
    }*/
    /*let frontend_image = await ImageChunk.from_blob((await gamefile.get_chunk_file(0x214)).get_chunk_of_type(0x4).contents);
    console.log(swizzle_4(384, 1936, frontend_image.width/32) / 128);
    console.log(deswizzle_32(2017*64, frontend_image.width/64));
    console.log(frontend_image.locations[271]);
    writeFileSync("font.data", frontend_image.export_data(frontend_image.locations[110]));
    writeFileSync("font.data.pal", frontend_image.export_data(frontend_image.locations[271], false, true));

    writeFileSync("frontend.data", frontend_image.data);
    console.log(frontend_image.width);
    console.log(frontend_image.height);*/
    let drackon_ridge = await gamefile.get_chunk_file(0x37D);
    let drackon_materials = await MaterialsChunk.from_blob(drackon_ridge.get_chunk_of_type(2).contents);
    let drackon_image3 = drackon_ridge.get_chunk_of_type(4, 3);
    let water_material = drackon_materials.materials[13];
    let map_file = await gamefile.get_chunk_file(0x512);
    map_file.chunks = map_file.chunks.filter(c => {
        return [
            ChunkType.Header,
            ChunkType.Materials,
            ChunkType.WorldModel,
            ChunkType.WorldInfo,
            ChunkType.Image,
            ChunkType.WorldGrid,
            ChunkType.Collision
        ].includes(c.type);
    });
    //let materials = await MaterialsChunk.from_blob(skybox_file.get_chunk_of_type(2).contents);
    //let models = await ChunkFile.from_blob(skybox_file.get_chunk_of_type(8).contents);
    let materials_chunk = map_file.get_chunk_of_type(2);
    let materials = await MaterialsChunk.from_blob(materials_chunk.contents);
    let images = await Promise.all(map_file.get_chunks_of_type(4).map(c => ImageChunk.from_blob(c.contents)));
    let mc_image = new ImageChunk(128, 128);
    images.push(mc_image);
    let extra_materials = [];
    for (let i = 1; i <= 27; i++) {
        let x = ((i - 1) % 8) * 16;
        let y = Math.floor((i - 1) / 8) * 16;
        let addr = swizzle_32(x, y, 2) / 64;
        console.log(addr);
        let reservation = {
            format: GsStorageFormat.PSMCT32,
            location: addr,
            height: 8,
            width: 8,
            is_clut: false
        };
        mc_image.import_data(reservation, PNG.sync.read(readFileSync("various_files/minecraft" + i + ".png")).data, false, true);
        let material = new Material();
        material.texture_file = images.length - 1;
        let pass = new MaterialPass();
        material.passes.push(pass);
        pass.texture_format = GsStorageFormat.PSMCT32;
        pass.texture_buffer_width = 128;
        pass.texture_log_width = 4;
        pass.texture_log_height = 4;
        pass.texture_location = [addr];
        pass.mag_filter = GsFilter.NEAREST;
        console.log(`minecraft${i} = ${materials.materials.length}`);
        materials.materials.push(material);
        if (i == 15 || i == 16) {
            let material = new Material();
            material.texture_file = images.length - 1;
            let pass = new MaterialPass();
            material.passes.push(pass);
            pass.texture_format = GsStorageFormat.PSMCT32;
            pass.texture_buffer_width = 128;
            pass.texture_log_width = 4;
            pass.texture_log_height = 4;
            pass.texture_location = [addr];
            pass.mag_filter = GsFilter.NEAREST;
            pass.alpha_blend_a = GsColorParam.RgbDest;
            pass.alpha_blend_b = GsColorParam.RgbDest;
            pass.alpha_blend_d = GsColorParam.RgbSource;
            pass.alpha_test_on = false;
            extra_materials.push(material);
        }
    }
    let clown_reservation = {
        format: GsStorageFormat.PSMCT32,
        location: swizzle_32(0, 64, 2) / 64,
        height: 32,
        width: 32,
        is_clut: false
    };
    mc_image.import_data(clown_reservation, PNG.sync.read(readFileSync("various_files/clown.png")).data, false, true);
    materials.materials.push(...extra_materials);
    map_file.chunks.push(new Chunk(mc_image.to_blob(), ChunkType.Image));
    materials.num_texture_files++;
    water_material.texture_file = images.length;
    console.log("FORMAT: " + GsStorageFormat[water_material.passes[0].texture_format]);
    materials.materials.push(water_material);
    images.push(await ImageChunk.from_blob(drackon_image3.contents));
    map_file.chunks.push(drackon_image3);
    materials.num_texture_files++;
    let clown_material = new Material();
    clown_material.texture_file = images.length - 2;
    let clown_pass = new MaterialPass();
    clown_material.passes.push(clown_pass);
    clown_pass.texture_format = GsStorageFormat.PSMCT32;
    clown_pass.texture_buffer_width = 128;
    clown_pass.texture_log_width = 5;
    clown_pass.texture_log_height = 5;
    clown_pass.texture_location = [clown_reservation.location];
    clown_pass.mag_filter = GsFilter.NEAREST;
    materials.materials.push(clown_material);
    //for(let i = 1; i < materials.materials.length; i++) {
    //	materials.materials[i] = materials.materials[0];
    //}
    //let model_chunk = map_file.get_chunk_of_type(3);
    //console.log("Before: " + (model_chunk.contents.size / 1000000).toFixed(2) + "MB");
    //let model = await ModelChunk.from_blob(model_chunk.contents);
    //writeFileSync("debug_area.glb", export_gltf(model, materials, images));
    //console.log(replaced_model);
    //model_chunk.contents = replaced_model.to_blob();
    //console.log("After: " + (model_chunk.contents.size / 1000000).toFixed(2) + "MB");
    let collision_chunk = map_file.get_chunk_of_type(5);
    //let collision = await CollisionChunk.from_blob(collision_chunk.contents);
    let model_chunk = map_file.get_chunk_of_type(3);
    //let model = await ModelChunk.from_blob(model_chunk.contents);
    //for(let material of materials.materials) material.passes[0].scroll_rate_x = 1/60;
    materials_chunk.contents = materials.to_blob();
    let replaced_model = import_gltf(readFileSync("various_files/minecraft_imp.glb").buffer, materials);
    model_chunk.contents = replaced_model.model.to_blob();
    //if(replaced_model.collision) collision_chunk.contents = replaced_model.collision.to_blob();
    //let grid_chunk = map_file.get_chunk_of_type(ChunkType.WorldGrid);
    //let grid = await GridChunk.from_blob(grid_chunk.contents);
    //debugger;
    //writeFileSync("tors_farm_collision.glb", export_gltf(materials, images, {collision}));
    //collision.objects[110].outer_grid_width = 0;
    //collision.objects[110].outer_grid_height = 0;
    //collision.objects.length = 1;
    //collision.objects = collision.objects.filter(i => Math.random() <= 0.5);
    //collision_chunk.contents = collision.to_blob();
    //grid_chunk.contents = grid.to_blob();
    writeFileSync("various_files/minecraft_imp_reexport.glb", export_gltf(materials, images, replaced_model));
    await map_file.rebuild_grid();
    console.log((map_file.to_blob().size / 1000000) + "MB");
    gamefile.replace_chunk_file(0x512, map_file);
    //gamefile.replace_chunk_file(0x512, map_file);
    writeFileSync("various_files/debug_area.dat", new Uint8Array(await gamefile.get_file(0x512).arrayBuffer()));
    //for(let material of materials.materials)console.log(material);
    //console.log(materials);
    //console.log(JSON.stringify((await ImageChunk.from_blob((await gamefile.get_chunk_file(0x214)).get_chunk_of_type(0x4).contents)).locations));
    //console.log(JSON.parse((await )))
    if (JSON.stringify(gamefile.sector_map) != JSON.stringify(sector_map))
        console.log("Sector map has changed.");
    if (!process.argv.includes("output-full"))
        return;
    console.log("Writing file...");
    //Readable.from(gamefile.blob.stream()).pipe(createWriteStream("GAMEFILE.DAT"));
    //return;
    let modified_exe = await export_sector_map(exe_blob, gamefile.sector_map);
    //Readable.from(modified_exe.stream()).pipe(createWriteStream("SCUS_973.67"));
    let modified_iso = new Blob([
        iso_blob.slice(0, 390 * 2048),
        modified_exe,
        iso_blob.slice(390 * 2048 + modified_exe.size, 885373 * 2048),
        gamefile.blob,
        iso_blob.slice(885373 * 2048 + gamefile.blob.size)
    ]);
    //Readable.from(modified_iso.stream()).pipe(createWriteStream("darkestfaerie-nmt.iso"));
    //let filename = "D:\\dev\\ps2\\PS2SMB\\DVD\\SCUS_973.67.df mod test.iso";
    let filename = "darkestfaerie-nmt.iso";
    console.log();
    let stream = Readable.from(modified_iso.stream()).pipe(createWriteStream(filename));
    let interval = setInterval(() => {
        let percent = stream.bytesWritten / modified_iso.size;
        process.stdout.write("\r" + (percent * 100).toFixed(1) + "%     ");
    }, 100);
    stream.on('finish', () => {
        clearInterval(interval);
    });
})();
function add_lower(num) {
    return num & 0xFFFF;
}
function add_upper(num) {
    return (num >>> 16) + (num & 0x8000 ? 1 : 0);
}
